@@ -1,8 +1,10 @@
 #!/usr/bin/env python
 
-import md5, os, stat
+import anydbm, md5, os, stat, struct
 
-from pyutil import mathutil
+from cPickle import dumps, loads
+
+from pyutil import fileutil, mathutil
 
 READSIZE=8192
 
@@ -14,18 +16,20 @@ def md5it(f):
         d = f.read(READSIZE)
     return h.hexdigest()
 
-def dffem(VERBOSITY, IGNOREDIRS, HARDLINKEM, DELETEEM, MINSIZE, PROFILES, DIRS, next_multiple=mathutil.next_multiple):
+def dffem(VERBOSITY, IGNOREDIRS, HARDLINKEM, DELETEEM, MINSIZE, PROFILES, DIRS):
+    # d1 = { size: {(dev, ino): [fname]} }
+    d1 = anydbm.open("dupfilefind_d2", "n")
+    try:
+        return _dffem(VERBOSITY, IGNOREDIRS, HARDLINKEM, DELETEEM, MINSIZE, PROFILES, DIRS, d1)
+    finally:
+        d1.close()
+        fileutil.remove_if_possible("dupfilefind_d1")
+
+def _dffem(VERBOSITY, IGNOREDIRS, HARDLINKEM, DELETEEM, MINSIZE, PROFILES, DIRS, d1, next_multiple=mathutil.next_multiple):
     assert not (HARDLINKEM and DELETEEM) # at most one of these
-    # d1 = { k: size, v: { k: (dev, ino,), v: [ fname ] } }
-    d1 = {}
 
-    visiteddirs = set() # to avoid symlink recursion
-
-    # First pass: build map from size, dev, inode number to filenames, and if PROFILES then print out the first 4 chars of the md5sum and size of each file.
+    # First pass: build map from size, dev, inode number to filenames.
     def visit(dirpath, dirnames, fnames, next_multiple=next_multiple):
-        if dirpath in visiteddirs:
-            return
-        visiteddirs.add(dirpath)
         for IGNOREDIR in IGNOREDIRS:
             if IGNOREDIR in dirnames:
                 dirnames.remove(IGNOREDIR)
@@ -36,13 +40,16 @@ def dffem(VERBOSITY, IGNOREDIRS, HARDLINKEM, DELETEEM, MINSIZE, PROFILES, DIRS,
             try:
                 thestat = os.stat(fullfname)
                 if not stat.S_ISLNK(thestat.st_mode) and thestat.st_size >= MINSIZE:
-                    d2 = d1.setdefault(thestat.st_size, {})
-                    if d2.has_key((thestat.st_dev, thestat.st_ino,)):
-                        l = d2[(thestat.st_dev, thestat.st_ino,)]
+                    d1k = struct.pack("@Q", thestat.st_size)
+                    if d1.has_key(d1k):
+                        d2 = loads(d1[d1k])
                     else:
-                        l = []
-                        d2[(thestat.st_dev, thestat.st_ino,)] = l
-                    l.append(fullfname)
+                        d2 = {}
+
+                    fnames = d2.get((thestat.st_dev, thestat.st_ino), [])
+                    fnames.append(fullfname)
+                    d2[(thestat.st_dev, thestat.st_ino)] = fnames
+                    d1[d1k] = dumps(d2)
             except EnvironmentError, le:
                 if VERBOSITY:
                     print "note: couldn't stat file named %s because %s" % (fullfname, le)
@@ -58,13 +65,15 @@ def dffem(VERBOSITY, IGNOREDIRS, HARDLINKEM, DELETEEM, MINSIZE, PROFILES, DIRS,
             visit(dirpath, dirnames, fnames)
 
     # Second pass: for any files which have identical size and different (dev, ino,), md5 that file.  Build map from md5 to { (dev, ino,): [ fname ] }.
-    # ... and print out any collisions (that is: same md5, different (dev, ino,)).
+    # ... and print out any duplicates (that is: same md5, different (dev, ino,)).
 
-    # d3 = { k: md5hash, v: { k: (dev, ino,), v: [ fname ] } }
-    d3 = {}
-
-    for (size, d2,) in d1.iteritems():
+    for (sizestr, d2str,) in d1.iteritems():
+        size = struct.unpack("@Q", sizestr)[0]
+        d2 = loads(d2str)
         if (len(d2) > 1) or (len(d2) == 1 and PROFILES):
+            # d3 = { k: md5hash, v: { k: (dev, ino,), v: [ fname ] } }
+            d3 = {}
+
             if VERBOSITY:
                 print "files with same sizes.  size: %s, fnames: %s" % (size, d2,)
             for ((dev, ino,), fnames,) in d2.iteritems():
